/*
Copyright (C) 2010 Filip 'RocketGuy' Juresa

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "../quakedef.h"
#include "ban.h"

//========================================================
// strlcat and strlcpy, from OpenBSD

/*
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*	$OpenBSD: strlcat.c,v 1.11 2003/06/17 21:56:24 millert Exp $	*/
/*	$OpenBSD: strlcpy.c,v 1.8 2003/06/17 21:56:24 millert Exp $	*/


#ifndef HAVE_STRLCAT
size_t
strlcat(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}
#endif  // #ifndef HAVE_STRLCAT


#ifndef HAVE_STRLCPY
size_t
strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

#endif  // #ifndef HAVE_STRLCPY

banaddr_t *Ban_Alloc (void)
{
	banaddr_t *old = ban_list;

	if (num_bans == -1)
	{
		Con_Printf("Allocating memory for the ban list has failed, please restart Quake to try again\n");
		return NULL;
	}
	
	ban_list = (banaddr_t *)realloc(ban_list, ++num_bans * sizeof(banaddr_t));
	if (!ban_list)
	{
		Con_Printf("Could not allocate memory for the ban_list\n");
		num_bans = -1;
		free (old);
		return NULL;
	}
	else
		return ban_list + num_bans-1;
}

void Ban_Cleanup (unsigned int num)
{
	int i;
	banaddr_t *old = ban_list;
	
	for (i = num; i < num_bans-1; i++)
	{
		ban_list[i].addr = ban_list[i+1].addr;
		ban_list[i].mask = ban_list[i+1].mask;
	}
	num_bans--;
	ban_list = realloc(ban_list, num_bans*sizeof(banaddr_t));
	
	if (!ban_list && num_bans)
	{
		Con_Printf("Could not allocate memory for the ban_list\n");
		free (old);
		num_bans = -1;
	}
}


#ifdef BANLIST_XML
void Ban_LoadFromFile (void)
{
	FILE		*f;
	char		*buffer, *banstr, ban[32], addr_str[16], mask_str[16];
	size_t		len, start, end;
	banaddr_t	*banaddr;

	f = fopen(va("%s/banlist.xml",com_gamedir), "r");

	if (!f)
	{
		Con_Printf("banlist.xml does not exist\n");
		return;
	}
	else
	{
		fseek(f, 0, SEEK_END);
		len = ftell(f);
		rewind(f);
		buffer = (char *)malloc(sizeof(char)*(len+1));
		fread(buffer, 1, len, f);
		fclose(f);

		buffer[len]= '\0';
		snprintf(&ban[0], 32, "<ban1>");
		
		banstr = strchr(buffer, '\n');
		
		if (banstr)
		{
			banstr++;
			if (strncmp(banstr, "<?xml", 5))
			{
				Con_Printf("banlist.xml is corrupt\n");
				free(buffer);
				return;
			}
		}
		else
		{
			Con_Printf("banlist.xml is empty\n");
			free(buffer);
			return;
		}
		banstr = strstr(buffer, ban);

		while(banstr)
		{
			banaddr = Ban_Alloc();

			banstr += strlen(ban);

			while (isspace(*banstr))
				banstr++;

			banstr += 6; //<addr>
			start = strlen(banstr);
			banstr = strstr(banstr, "</addr>");
			end = strlen(banstr);

			if (start-end > 15)
			{
				Con_Printf("Invalid entry #%u in banlist.xml, address is longer than 15\n", num_bans);
				continue;
			}

			banstr -= start-end;
			strlcpy(addr_str, banstr, start-end);

			banstr = strstr(banstr, "<mask>");

			banstr += 6;
			start = strlen(banstr);
			banstr = strstr(banstr, "</mask>");
			end = strlen(banstr);

			if (start-end > 15)
			{			
				Con_Printf("Invalid entry #%u in banlist.xml, subnet mask is longer than 15\n", num_bans);
				continue;
			}

			banstr -= start-end;
			strlcpy(mask_str, banstr, start-end);
			
			banaddr->addr = inet_addr(&addr_str[0]);
			banaddr->mask = inet_addr(&mask_str[0]);

			if (banaddr->addr == INADDR_NONE)
			{
				Con_Printf("Invalid entry #%u in banlist.xml\n");
				banaddr->addr = 0;
				banaddr->mask = 0;
			}

			snprintf(&ban[0], 32, "<ban%u>", num_bans+1);
			banstr = strstr(banstr, ban);
		}

		free(buffer);
		return;
	}
}

void Ban_WriteToFile (void)
{
	FILE	*f;
	char	addr[16];
	int		i;

	if (num_bans == -1) // means that the memory for the banlist could not be allocated, so don't overwrite the old one with blank
		return;

	f = fopen(va("%s/banlist.xml",com_gamedir), "w");
	if (!f)
	{
		Con_Printf("Couldn't open banlist.xml for writing\n");
		return;
	}
	
	fputs("<?-- This file was generated by Quake, don't edit it manually?>\n"
				"<?xml version=\"1.0\" encoding=\"UTF-8\"?>", f);
	for (i = 0; i < num_bans; i++)
	{
		strlcpy(&addr[0], inet_ntoa(*(struct in_addr *)&ban_list[i].addr), sizeof(addr));
		fprintf(f, "\n<ban%u>\n\t<addr>%s</addr>\n\t<mask>%s</mask>\n</ban%u>", i+1, addr, inet_ntoa(*(struct in_addr*)&ban_list[i].mask), i+1);
	}
	fclose(f);
}
#else
void Ban_LoadFromFile (void)
{
	Cbuf_AddText("exec banlist.cfg");
}
void Ban_WriteToFile (void)
{
	FILE	*f;
	char	addr[16];
	int		i;

	if (num_bans == -1) // means that the memory for the banlist could not be allocated, so don't overwrite the old one with blank
		return;

	f = fopen(va("%s/banlist.cfg",com_gamedir), "w");
	if (!f)
	{
		Con_Printf("Couldn't open banlist.cfg for writing\n");
		return;
	}

	for (i = 0; i < num_bans; i++)
	{
		strlcpy(&addr[0], inet_ntoa(*(struct in_addr *)&ban_list[i].addr), sizeof(addr));
		fprintf(f, "addip %s %s\n", addr, inet_ntoa(*(struct in_addr*)&ban_list[i].mask));
	}
	fclose(f);
}
#endif

void NET_Addip (unsigned long addr, unsigned long mask)
{
	int i;
	banaddr_t *b;

	for (i = 0; i < num_bans; i++)
	{
		if (ban_list[i].addr == addr)
		{
			if (ban_list[i].mask != mask)
			{
				Con_Printf("Entry #%i mask changed to %s\n", i+1, inet_ntoa(*(struct in_addr *)&mask));
				ban_list[i].mask = mask;
				return;
			}
			else
			{
			Con_Printf("%s is already on the banlist\n", inet_ntoa(*(struct in_addr *)&addr));
			return;
			}
		}
	}

	b = Ban_Alloc();
	if (b)
	{
		b->addr = addr;
		b->mask = mask;
	}
}

void NET_Addip_f (void)
{
	banaddr_t	*b;
	int		i;

	if (inet_addr(Cmd_Argv(1)) == INADDR_NONE)
	{
		Con_Printf("Invalid address\n");
		return;
	}

	switch (Cmd_Argc ())
	{
		case 2:
			NET_Addip(inet_addr(Cmd_Argv(1)), 0xFFFFFFFF);
			break;

		case 3:
			NET_Addip(inet_addr(Cmd_Argv(1)), inet_addr(Cmd_Argv(2)));
			break;

		default:
			Con_Printf("Usage: addip [ip]\n or addip [ip] [subnet mask]\n");
			break;
	}
}

void NET_Unban_f (void)
{
	unsigned long addr;
	int i;

	switch (Cmd_Argc())
	{
		case 2:
			addr = inet_addr(Cmd_Argv(1));
			if (addr == INADDR_NONE)
			{
				Con_Printf("Usage: unban [ip] or unban # [number]\n");
				break;
			}
			for (i = 0; i < num_bans; i++)
			{
				if (ban_list[i].addr == addr)
					break;
			}
			if (i == num_bans-1)
				Con_Printf("%s is not on the banlist\n", Cmd_Argv(1));
			else
				Ban_Cleanup(i);
			break;
	
		case 3:
			i = atoi(Cmd_Argv(2))-1;
			if (Cmd_Argv(1)[0] != '#' || i < 0)
			{
				Con_Printf("Usage: unban [ip] or unban # [number]\n");
				break;
			}
			if (i+1 > num_bans)
			{
				Con_Printf("%s is out of range\n", Cmd_Argv(2));
				break;
			}
			Ban_Cleanup(i);
			break;

		default:
			Con_Printf("Usage: unban [ip] or unban # [number]\n");
			break;
	}
}

void NET_Banlist_f (void)
{
	char	addr[16];
	int i;

	if (!num_bans)
	{
		Con_Printf("The banlist is empty\n");
		return;
	}

	Con_Printf("Banlist:\n");
	for (i = 0; i < num_bans; i++)
	{
		strlcpy(&addr[0], inet_ntoa(*(struct in_addr *)&ban_list[i].addr), sizeof(addr));
		Con_Printf("#%i: %s [%s]\n", i+1, addr, inet_ntoa(*(struct in_addr *)&ban_list[i].mask));
	}
}

void Host_Kickban_f (void)
{
	char		*who;
	char		addr[32], *sc;
	char		*message = NULL;
	client_t	*save;
	int			i, j, k;
	qboolean	byNumber = false;

	if (cmd_source == src_command)
	{
		if (!sv.active)
		{
			Cmd_ForwardToServer ();
			return;
		}
	}
	else if (pr_global_struct->deathmatch && !host_client->privileged)
		return;

	save = host_client;

	if (Cmd_Argc() > 2 && Q_strcmp(Cmd_Argv(1), "#") == 0)
	{
		i = Q_atof(Cmd_Argv(2)) - 1;
		if (i < 0 || i >= svs.maxclients)
			return;
		if (!svs.clients[i].active)
			return;
		host_client = &svs.clients[i];
		byNumber = true;
	}
	else
	{
		for (i = 0, host_client = svs.clients; i < svs.maxclients; i++, host_client++)
		{
			if (!host_client->active)
				continue;
			if (Q_strcasecmp(host_client->name, Cmd_Argv(1)) == 0)
				break;
		}
	}

	if (i < svs.maxclients)
	{
		if (cmd_source == src_command)
			if (cls.state == ca_dedicated)
				who = "Console";
			else
				who = cl_name.string;
		else
			who = save->name;

		// can't kick yourself!
		if (host_client == save)
			return;

		if (Cmd_Argc() > 2)
		{
			message = COM_Parse(Cmd_Args());
			if (byNumber)
			{
				message++;							// skip the #
				while (*message == ' ')				// skip white space
					message++;
				message += Q_strlen(Cmd_Argv(2));	// skip the number
			}
			while (*message && *message == ' ')
				message++;
		}
		if (*message)
			SV_ClientPrintf ("Kicked and banned: %s\n", message);
		else
			SV_ClientPrintf ("Kicked and banned\n");
		
		// strip the port number
		sc = strchr(&host_client->netconnection->address[0], ':');
		j = strlen(sc);
		j = strlen(&host_client->netconnection->address[0])-j;
		//strlcpy(&addr[0], &host_client->netconnection->address[0], min(j+1, sizeof (addr)));
		strlcpy(&addr[0], &host_client->netconnection->address[0], (j+1 < sizeof (addr) ? j+1 : sizeof(addr)));
		
		NET_Addip(inet_addr(addr), 0xFFFFFFFF);
		SV_DropClient (false);

		// Drop any other clients with this address
		for (k = 0; k < svs.maxclients; k++)
		{
			if (!svs.clients[k].active)
				continue;

			if (!strncmp(svs.clients[k].netconnection->address, addr, j))
			{
				host_client = &svs.clients[k];
				if (*message)
					SV_ClientPrintf ("Kicked and banned: %s\n", message);
				else
					SV_ClientPrintf ("Kicked and banned\n");

				SV_DropClient(false);
			}
		}
	}
	host_client = save;
}

void Ban_Init (void)
{
	num_bans = 0;
	Ban_LoadFromFile();
	
	Cmd_AddCommand ("addip", NET_Addip_f);
	Cmd_AddCommand ("banlist", NET_Banlist_f);
	Cmd_AddCommand ("unban", NET_Unban_f);
	Cmd_AddCommand ("kickban", Host_Kickban_f);
};

void Ban_Shutdown (void)
{
	Ban_WriteToFile ();
	free (ban_list);
}
